"use strict";
// Copyright (c) Jupyter Development Team.
// Distributed under the terms of the Modified BSD License.
Object.defineProperty(exports, "__esModule", { value: true });
var signaling_1 = require("@phosphor/signaling");
var coreutils_1 = require("@jupyterlab/coreutils");
var observables_1 = require("@jupyterlab/observables");
var services_1 = require("@jupyterlab/services");
var github_1 = require("./github");
var base64js = require("base64-js");
/**
 * A Contents.IDrive implementation that serves as a read-only
 * view onto GitHub repositories.
 */
var GitHubDrive = /** @class */ (function () {
    /**
     * Construct a new drive object.
     *
     * @param options - The options used to initialize the object.
     */
    function GitHubDrive(registry) {
        var _this = this;
        this._validUser = false;
        this._isDisposed = false;
        this._fileChanged = new signaling_1.Signal(this);
        this._serverSettings = services_1.ServerConnection.makeSettings();
        this._fileTypeForPath = function (path) {
            var types = registry.getFileTypesForPath(path);
            return types.length === 0 ?
                registry.getFileType('text') :
                types[0];
        };
        // Test an api request to the notebook server
        // to see if the server proxy is installed.
        // If so, use that. If not, warn the user and
        // use the client-side implementation.
        this._useProxy = new Promise(function (resolve) {
            github_1.proxiedApiRequest('', _this._serverSettings).then(function () {
                resolve(true);
            }).catch(function () {
                console.warn('The JupyterLab GitHub server extension appears ' +
                    'to be missing. If you do not install it with application ' +
                    'credentials, you are likely to be rate limited by GitHub ' +
                    'very quickly');
                resolve(false);
            });
        });
        // Initialize the rate-limited observable.
        this.rateLimitedState = new observables_1.ObservableValue(false);
    }
    Object.defineProperty(GitHubDrive.prototype, "name", {
        /**
         * The name of the drive.
         */
        get: function () {
            return 'GitHub';
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(GitHubDrive.prototype, "validUser", {
        /**
         * State for whether the user is valid.
         */
        get: function () {
            return this._validUser;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(GitHubDrive.prototype, "fileChanged", {
        /**
         * A signal emitted when a file operation takes place.
         */
        get: function () {
            return this._fileChanged;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(GitHubDrive.prototype, "isDisposed", {
        /**
         * Test whether the manager has been disposed.
         */
        get: function () {
            return this._isDisposed;
        },
        enumerable: true,
        configurable: true
    });
    /**h
     * Dispose of the resources held by the manager.
     */
    GitHubDrive.prototype.dispose = function () {
        if (this.isDisposed) {
            return;
        }
        this._isDisposed = true;
        signaling_1.Signal.clearData(this);
    };
    Object.defineProperty(GitHubDrive.prototype, "baseURL", {
        /**
         * Get the base url of the manager.
         */
        get: function () {
            return github_1.GITHUB_API;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Get a file or directory.
     *
     * @param path: The path to the file.
     *
     * @param options: The options used to fetch the file.
     *
     * @returns A promise which resolves with the file content.
     */
    GitHubDrive.prototype.get = function (path, options) {
        var _this = this;
        var resource = parsePath(path);
        // If the org has not been set, return an empty directory
        // placeholder.
        if (resource.user === '') {
            this._validUser = false;
            return Promise.resolve(Private.dummyDirectory);
        }
        // If the org has been set and the path is empty, list
        // the repositories for the org.
        if (resource.user && !resource.repository) {
            return this._listRepos(resource.user);
        }
        // Otherwise identify the repository and get the contents of the
        // appropriate resource.
        var apiPath = coreutils_1.URLExt.encodeParts(coreutils_1.URLExt.join('repos', resource.user, resource.repository, 'contents', resource.path));
        return this._apiRequest(apiPath).then(function (contents) {
            // Set the states
            _this._validUser = true;
            if (_this.rateLimitedState.get() !== false) {
                _this.rateLimitedState.set(false);
            }
            return Private.gitHubContentsToJupyterContents(path, contents, _this._fileTypeForPath);
        }).catch(function (err) {
            if (err.response.status === 404) {
                console.warn('GitHub: cannot find org/repo. ' +
                    'Perhaps you misspelled something?');
                _this._validUser = false;
                return Private.dummyDirectory;
            }
            else if (err.response.status === 403 &&
                err.message.indexOf('rate limit') !== -1) {
                if (_this.rateLimitedState.get() !== true) {
                    _this.rateLimitedState.set(true);
                }
                console.error(err.message);
                return Promise.reject(err);
            }
            else if (err.response.status === 403 &&
                err.message.indexOf('blob') !== -1) {
                // Set the states
                _this._validUser = true;
                if (_this.rateLimitedState.get() !== false) {
                    _this.rateLimitedState.set(false);
                }
                return _this._getBlob(path);
            }
            else {
                console.error(err.message);
                return Promise.reject(err);
            }
        });
    };
    /**
     * Get an encoded download url given a file path.
     *
     * @param path - An absolute POSIX file path on the server.
     *
     * #### Notes
     * It is expected that the path contains no relative paths,
     * use [[ContentsManager.getAbsolutePath]] to get an absolute
     * path if necessary.
     */
    GitHubDrive.prototype.getDownloadUrl = function (path) {
        // Parse the path into user/repo/path
        var resource = parsePath(path);
        // Error if the user has not been set
        if (!resource.user) {
            return Promise.reject('GitHub: no active organization');
        }
        // Error if there is no path.
        if (!resource.path) {
            return Promise.reject('GitHub: No file selected');
        }
        // Otherwise identify the repository and get the url of the
        // appropriate resource.
        var dirname = coreutils_1.PathExt.dirname(resource.path);
        var dirApiPath = coreutils_1.URLExt.encodeParts(coreutils_1.URLExt.join('repos', resource.user, resource.repository, 'contents', dirname));
        return this._apiRequest(dirApiPath).then(function (dirContents) {
            for (var _i = 0, dirContents_1 = dirContents; _i < dirContents_1.length; _i++) {
                var item = dirContents_1[_i];
                if (item.path === resource.path) {
                    return item.download_url;
                }
            }
            throw Private.makeError(404, "Cannot find file at " + resource.path);
        });
    };
    /**
     * Create a new untitled file or directory in the specified directory path.
     *
     * @param options: The options used to create the file.
     *
     * @returns A promise which resolves with the created file content when the
     *    file is created.
     */
    GitHubDrive.prototype.newUntitled = function (options) {
        if (options === void 0) { options = {}; }
        return Promise.reject('Repository is read only');
    };
    /**
     * Delete a file.
     *
     * @param path - The path to the file.
     *
     * @returns A promise which resolves when the file is deleted.
     */
    GitHubDrive.prototype.delete = function (path) {
        return Promise.reject('Repository is read only');
    };
    /**
     * Rename a file or directory.
     *
     * @param path - The original file path.
     *
     * @param newPath - The new file path.
     *
     * @returns A promise which resolves with the new file contents model when
     *   the file is renamed.
     */
    GitHubDrive.prototype.rename = function (path, newPath) {
        return Promise.reject('Repository is read only');
    };
    /**
     * Save a file.
     *
     * @param path - The desired file path.
     *
     * @param options - Optional overrides to the model.
     *
     * @returns A promise which resolves with the file content model when the
     *   file is saved.
     */
    GitHubDrive.prototype.save = function (path, options) {
        return Promise.reject('Repository is read only');
    };
    /**
     * Copy a file into a given directory.
     *
     * @param path - The original file path.
     *
     * @param toDir - The destination directory path.
     *
     * @returns A promise which resolves with the new contents model when the
     *  file is copied.
     */
    GitHubDrive.prototype.copy = function (fromFile, toDir) {
        return Promise.reject('Repository is read only');
    };
    /**
     * Create a checkpoint for a file.
     *
     * @param path - The path of the file.
     *
     * @returns A promise which resolves with the new checkpoint model when the
     *   checkpoint is created.
     */
    GitHubDrive.prototype.createCheckpoint = function (path) {
        return Promise.reject('Repository is read only');
    };
    /**
     * List available checkpoints for a file.
     *
     * @param path - The path of the file.
     *
     * @returns A promise which resolves with a list of checkpoint models for
     *    the file.
     */
    GitHubDrive.prototype.listCheckpoints = function (path) {
        return Promise.resolve([]);
    };
    /**
     * Restore a file to a known checkpoint state.
     *
     * @param path - The path of the file.
     *
     * @param checkpointID - The id of the checkpoint to restore.
     *
     * @returns A promise which resolves when the checkpoint is restored.
     */
    GitHubDrive.prototype.restoreCheckpoint = function (path, checkpointID) {
        return Promise.reject('Repository is read only');
    };
    /**
     * Delete a checkpoint for a file.
     *
     * @param path - The path of the file.
     *
     * @param checkpointID - The id of the checkpoint to delete.
     *
     * @returns A promise which resolves when the checkpoint is deleted.
     */
    GitHubDrive.prototype.deleteCheckpoint = function (path, checkpointID) {
        return Promise.reject('Read only');
    };
    /**
     * If a file is too large (> 1Mb), we need to access it over the
     * GitHub Git Data API.
     */
    GitHubDrive.prototype._getBlob = function (path) {
        var _this = this;
        var blobData;
        // Get the contents of the parent directory so that we can
        // get the sha of the blob.
        var resource = parsePath(path);
        var dirname = coreutils_1.PathExt.dirname(resource.path);
        var dirApiPath = coreutils_1.URLExt.encodeParts(coreutils_1.URLExt.join('repos', resource.user, resource.repository, 'contents', dirname));
        return this._apiRequest(dirApiPath).then(function (dirContents) {
            for (var _i = 0, dirContents_2 = dirContents; _i < dirContents_2.length; _i++) {
                var item = dirContents_2[_i];
                if (item.path === resource.path) {
                    blobData = item;
                    return item.sha;
                }
            }
            throw Error('Cannot find sha for blob');
        }).then(function (sha) {
            // Once we have the sha, form the api url and make the request.
            var blobApiPath = coreutils_1.URLExt.encodeParts(coreutils_1.URLExt.join('repos', resource.user, resource.repository, 'git', 'blobs', sha));
            return _this._apiRequest(blobApiPath);
        }).then(function (blob) {
            // Convert the data to a Contents.IModel.
            blobData.content = blob.content;
            return Private.gitHubContentsToJupyterContents(path, blobData, _this._fileTypeForPath);
        });
    };
    /**
     * List the repositories for the currently active user.
     */
    GitHubDrive.prototype._listRepos = function (user) {
        var _this = this;
        // First, check if the `user` string is actually an org.
        // If will return with an error if not, and we can try
        // the user path.
        var apiPath = coreutils_1.URLExt.encodeParts(coreutils_1.URLExt.join('orgs', user, 'repos'));
        return this._apiRequest(apiPath).catch(function (err) {
            // If we can't find the org, it may be a user.
            if (err.response.status === 404) {
                // Check if it is the authenticated user.
                return _this._apiRequest('user').then(function (currentUser) {
                    var reposPath;
                    // If we are looking at the currently authenticated user,
                    // get all the repositories they own, which includes private ones.
                    if (currentUser.login === user) {
                        reposPath = coreutils_1.URLExt.join('user', 'repos', '?type=owner');
                    }
                    else {
                        reposPath = coreutils_1.URLExt.encodeParts(coreutils_1.URLExt.join('users', user, 'repos'));
                    }
                    return _this._apiRequest(reposPath);
                }).catch(function (err) {
                    // If there is no authenticated user, return the public
                    // users api path.
                    if (err.response.status === 401) {
                        var reposPath = coreutils_1.URLExt.encodeParts(coreutils_1.URLExt.join('users', user, 'repos'));
                        return _this._apiRequest(reposPath);
                    }
                    throw err;
                });
            }
            throw err;
        }).then(function (repos) {
            // Set the states
            _this._validUser = true;
            if (_this.rateLimitedState.get() !== false) {
                _this.rateLimitedState.set(false);
            }
            return Private.reposToDirectory(repos);
        }).catch(function (err) {
            if (err.response.status === 403 &&
                err.message.indexOf('rate limit') !== -1) {
                if (_this.rateLimitedState.get() !== true) {
                    _this.rateLimitedState.set(true);
                }
            }
            else {
                console.warn('GitHub: cannot find user. ' +
                    'Perhaps you misspelled something?');
                _this._validUser = false;
            }
            return Private.dummyDirectory;
        });
    };
    /**
     * Determine whether to make the call via the
     * notebook server proxy or not.
     */
    GitHubDrive.prototype._apiRequest = function (apiPath) {
        var _this = this;
        return this._useProxy.then(function (result) {
            if (result === true) {
                return github_1.proxiedApiRequest(apiPath, _this._serverSettings);
            }
            else {
                return github_1.browserApiRequest(apiPath);
            }
        });
    };
    return GitHubDrive;
}());
exports.GitHubDrive = GitHubDrive;
/**
 * Parse a path into a IGitHubResource.
 */
function parsePath(path) {
    var parts = path.split('/');
    var user = parts.length > 0 ? parts[0] : '';
    var repository = parts.length > 1 ? parts[1] : '';
    var repoPath = parts.length > 2 ? coreutils_1.URLExt.join.apply(coreutils_1.URLExt, parts.slice(2)) : '';
    return { user: user, repository: repository, path: repoPath };
}
exports.parsePath = parsePath;
/**
 * Private namespace for utility functions.
 */
var Private;
(function (Private) {
    /**
     * A dummy contents model indicating an invalid or
     * nonexistent repository.
     */
    Private.dummyDirectory = {
        type: 'directory',
        path: '',
        name: '',
        format: 'json',
        content: [],
        created: '',
        writable: false,
        last_modified: '',
        mimetype: '',
    };
    /**
     * Given a JSON GitHubContents object returned by the GitHub API v3,
     * convert it to the Jupyter Contents.IModel.
     *
     * @param path - the path to the contents model in the repository.
     *
     * @param contents - the GitHubContents object.
     *
     * @param fileTypeForPath - a function that, given a path, returns
     *   a DocumentRegistry.IFileType, used by JupyterLab to identify different
     *   openers, icons, etc.
     *
     * @returns a Contents.IModel object.
     */
    function gitHubContentsToJupyterContents(path, contents, fileTypeForPath) {
        if (Array.isArray(contents)) {
            // If we have an array, it is a directory of GitHubContents.
            // Iterate over that and convert all of the items in the array/
            return {
                name: coreutils_1.PathExt.basename(path),
                path: path,
                format: 'json',
                type: 'directory',
                writable: false,
                created: '',
                last_modified: '',
                mimetype: '',
                content: contents.map(function (c) {
                    return gitHubContentsToJupyterContents(coreutils_1.PathExt.join(path, c.name), c, fileTypeForPath);
                })
            };
        }
        else if (contents.type === 'file' || contents.type === 'symlink') {
            // If it is a file or blob, convert to a file
            var fileType = fileTypeForPath(path);
            var fileContents = contents.content;
            var content = void 0;
            switch (fileType.fileFormat) {
                case 'text':
                    content = fileContents !== undefined ?
                        Private.b64DecodeUTF8(fileContents) :
                        null;
                    break;
                case 'base64':
                    content = fileContents !== undefined ? fileContents : null;
                    break;
                case 'json':
                    content = fileContents !== undefined ?
                        JSON.parse(Private.b64DecodeUTF8(fileContents)) :
                        null;
                    break;
                default:
                    throw new Error("Unexpected file format: " + fileType.fileFormat);
            }
            return {
                name: coreutils_1.PathExt.basename(path),
                path: path,
                format: fileType.fileFormat,
                type: 'file',
                created: '',
                writable: false,
                last_modified: '',
                mimetype: fileType.mimeTypes[0],
                content: content
            };
        }
        else if (contents.type === 'dir') {
            // If it is a directory, convert to that.
            return {
                name: coreutils_1.PathExt.basename(path),
                path: path,
                format: 'json',
                type: 'directory',
                created: '',
                writable: false,
                last_modified: '',
                mimetype: '',
                content: null
            };
        }
        else if (contents.type === 'submodule') {
            // If it is a submodule, throw an error, since we cannot
            // GET submodules at the moment. NOTE: due to a bug in the GithHub
            // API, the `type` for submodules in a directory listing is incorrectly
            // reported as `file`: https://github.com/github/developer.github.com/commit/1b329b04cece9f3087faa7b1e0382317a9b93490
            // This means submodules will show up in the listing, but we still should not
            // open them.
            throw makeError(400, "Cannot open \"" + contents.name + "\" because it is a submodule");
        }
        else {
            throw makeError(500, "\"" + contents.name + "\" has and unexpected type: " + contents.type);
        }
    }
    Private.gitHubContentsToJupyterContents = gitHubContentsToJupyterContents;
    /**
     * Given an array of JSON GitHubRepo objects returned by the GitHub API v3,
     * convert it to the Jupyter Contents.IModel conforming to a directory of
     * those repositories.
     *
     * @param repo - the GitHubRepo object.
     *
     * @returns a Contents.IModel object.
     */
    function reposToDirectory(repos) {
        // If it is a directory, convert to that.
        var content = repos.map(function (repo) {
            return {
                name: repo.name,
                path: repo.name,
                format: 'json',
                type: 'directory',
                created: '',
                writable: false,
                last_modified: '',
                mimetype: '',
                content: null
            };
        });
        return {
            name: '',
            path: '',
            format: 'json',
            type: 'directory',
            created: '',
            last_modified: '',
            writable: false,
            mimetype: '',
            content: content
        };
    }
    Private.reposToDirectory = reposToDirectory;
    /**
     * Wrap an API error in a hacked-together error object
     * masquerading as an `ServerConnection.ResponseError`.
     */
    function makeError(code, message) {
        var response = new Response(message, { status: code, statusText: message });
        return new services_1.ServerConnection.ResponseError(response, message);
    }
    Private.makeError = makeError;
    /**
     * Decoder from bytes to UTF-8.
     */
    var decoder = new TextDecoder('utf8');
    /**
     * Decode a base-64 encoded string into unicode.
     *
     * See https://developer.mozilla.org/en-US/docs/Web/API/WindowBase64/Base64_encoding_and_decoding#Solution_2_%E2%80%93_rewrite_the_DOMs_atob()_and_btoa()_using_JavaScript's_TypedArrays_and_UTF-8
     */
    function b64DecodeUTF8(str) {
        var bytes = base64js.toByteArray(str.replace(/\n/g, ''));
        return decoder.decode(bytes);
    }
    Private.b64DecodeUTF8 = b64DecodeUTF8;
})(Private || (Private = {}));
